;==============================================================================
;
;                             BIOS for NASCOM BASIC
;
; Supports 320x240 TFT with ili9325 controller and keyboard.
;
; Copyright Oliver Urbann 2013
; Feel free to use this code in your own non-commercial projects.
; For commercial use please contact the author.
;
; http://urbann.de.vu
;
; This project is based on the SBC project by Grant Searle and contains parts 
; of his code:
; http://searle.hostei.com/grant/index.html
;
;==============================================================================

#DEFINE TFTCMD(cmd,dat) \ LD BC, cmd \ CALL WRCMD \ LD BC, dat \ CALL WRDATA
#DEFINE SETCOLOR(col) \ PUSH BC \ LD BC, col \ LD (TXTCOL), BC \ POP BC

LCDCTRL     .EQU      60h
LCDDATA     .EQU      40h
REGISTER    .EQU      0
KEYREAD     .EQU      80h

CTRLC       .EQU      03H             ; Control "C"
CTRLG       .EQU      07H             ; Control "G"
BKSP        .EQU      08H             ; Back space
LF          .EQU      0AH             ; Line feed
CS          .EQU      0CH             ; Clear screen
CR          .EQU      0DH             ; Carriage return
CTRLO       .EQU      0FH             ; Control "O"
CTRLQ       .EQU      11H             ; Control "Q"
CTRLR       .EQU      12H             ; Control "R"
CTRLS       .EQU      13H             ; Control "S"
CTRLU       .EQU      15H             ; Control "U"
ESC         .EQU      1BH             ; Escape
DEL         .EQU      7FH             ; Delete

BLACK				.EQU		  0
BLUE				.EQU		  0FFH
GREEN				.EQU		  7E0H
RED					.EQU		  0F800H
WHITE				.EQU		  0FFFFH
YELLOW      .EQU      0FFF0H

BASE				.EQU		  4000H
CHARS				.EQU		  BASE+1
LINES				.EQU		  BASE
COORDX			.EQU		  BASE+3
COORDY			.EQU		  BASE+5
CRSRSTATE   .EQU		  BASE+7
TXTCOL      .EQU      BASE+9
SCRNOFFSET  .EQU      BASE+11

SCREENWIDTH .EQU      30
SCREENHEIGHT .EQU     40



TEMPSTACK  .EQU       BASE+0EDH

            .ORG      0
RST00       DI
            JP        INIT
            
            .ORG      0008H
RST08       JP        SAFEPRINT

            .ORG      0010H
RST10       JP        READKEYINIT

            .ORG      0018H
RST18       LD        A, 0
            CP        0
            RET
            
WELCOMEMSG:
.BYTE       "Z80 SBC ",0

KEYMAP:
.BYTE        "abcdefgh"
.BYTE        "ijklmnop"
.BYTE        "qrstuvwx"
.BYTE        "yz123456"
.BYTE        "7890,.-",CR
.BYTE        "<+[]{}?",BKSP
.BYTE        "        " 

SHIFTKEYMAP:
.BYTE        "ABCDEFGH"
.BYTE        "IJKLMNOP"
.BYTE        "QRSTUVWX"
.BYTE        "YZ!#~$%&"
.BYTE        "/()=;._",CR
.BYTE        ">*@|\\^",34,BKSP
.BYTE        "        " 

INIT:       LD        HL,TEMPSTACK    ; Temp stack
            LD        SP,HL           ; Set up a temporary stack    
            XOR       E               ; Empty E for key reading
            LD        A, 0            
            LD        (LINES), A      ; Line counter for LCD  
            LD        (CHARS), A      ; Characters in line
            LD        (CRSRSTATE), A  ; Counter for cursorchange
            LD        A, 1            ; First line at end of screen
            LD        (SCRNOFFSET), A
            LD        BC, GREEN
            LD        (TXTCOL), BC    ; Green is default text color
                  
;      *********** Init Display ***********
            LD        B, 50
            CALL      WAIT
            TFTCMD(   00FFH, 0001H)
            TFTCMD(   00F3H, 0008H)
            LD        BC, 00F3H
            CALL      WRCMD
            TFTCMD(   0001H, 0100H)
            TFTCMD(   0002H, 0700H)
            TFTCMD(   0003H, 1030H)
            TFTCMD(   0008H, 0207H)
            TFTCMD(   0009H, 0000H)
            TFTCMD(   0010H, 0000H)
            TFTCMD(   0011H, 0000H)
            TFTCMD(   0012H, 0000H)
            TFTCMD(   0013H, 0000H)
            LD        B, 50
            CALL      WAIT
            TFTCMD(   0010H, 1290H)
            LD        B, 50
            CALL      WAIT            
            TFTCMD(   0011H, 0227H)
            LD        B, 50
            CALL      WAIT            
            TFTCMD(   0012H, 001BH)
            TFTCMD(   0013H, 1100H)
            TFTCMD(   0029H, 0019H)
            LD        B, 50
            CALL      WAIT            
            TFTCMD(   0030H, 0000H)
            TFTCMD(   0031H, 0107H)
            TFTCMD(   0032H, 0000H)
            TFTCMD(   0035H, 0203H)
            TFTCMD(   0036H, 0402H)
            TFTCMD(   0037H, 0000H)
            TFTCMD(   0038H, 0207H)
            TFTCMD(   0039H, 0000H)
            TFTCMD(   003CH, 0203H)
            TFTCMD(   003DH, 0403H)
            TFTCMD(   0050H, 0000H)
            TFTCMD(   0051H, 00EFH)
            TFTCMD(   0052H, 0000H)
						CALL      CLEAR
            TFTCMD(   0053H, 013FH)
            TFTCMD(   0060H, 0A700H)
            TFTCMD(   0061H, 0003H)
            TFTCMD(   006AH, 0008H)
            TFTCMD(   0090H, 0029H)
            TFTCMD(   0007H, 0133H)
            LD        B, 50
            CALL      WAIT            
            LD        BC, 0022H
            CALL      WRCMD
            LD        B, 50
            CALL      WAIT
            
            IM        1
            LD        HL, WELCOMEMSG
            CALL      PRINT
            LD        BC, RED
            LD        (TXTCOL), BC
            LD        A, 'C'
            CALL      PRINTCHAR
            LD        BC, GREEN
            LD        (TXTCOL), BC
            LD        A, 'o'
            CALL      PRINTCHAR
            LD        BC, BLUE
            LD        (TXTCOL), BC
            LD        A, 'l'
            CALL      PRINTCHAR
            LD        BC, YELLOW
            LD        (TXTCOL), BC
            LD        A, 'o'
            CALL      PRINTCHAR
            LD        BC, RED
            LD        (TXTCOL), BC
            LD        A, 'r'
            CALL      PRINTCHAR
            LD        BC, GREEN
            LD        (TXTCOL), BC
            EI
            JP        $02000           ; Start BASIC
            HALT
            
WRCMD:      LD        A, B
            OUT       (LCDCTRL), A
            LD        A, C
            OUT       (LCDCTRL), A
            RET
            
WRDATA:     LD        A, B
            OUT       (LCDDATA), A
            LD        A, C
            OUT       (LCDDATA), A
            RET
		
WAIT:       
						LD        A, 100
            LD        C, B
SMALLWAIT:  DJNZ      SMALLWAIT
            LD        B, C
            DEC       A
            JP        NZ, SMALLWAIT
            RET        
            
PRINT:      LD        A,(HL)          ; Get character
            OR        A               ; Is it 0?
            RET       Z               ; Then RET
            PUSH      HL
            CALL      PRINTCHAR
            POP       HL
            INC       HL
            JP        PRINT
        
SAFEPRINT:	PUSH      AF
						PUSH      BC
						PUSH      DE
						PUSH      HL
						PUSH      IX
						PUSH      IY
						CALL      PRINTCHAR
						POP       IY
						POP       IX
						POP       HL
						POP       DE
						POP       BC
						POP       AF
						RET
            
PRINTCHAR:  CP        LF              ; Is char linefeed?
            JP        NZ, NONEWLINE   ; If not skip new line
            LD        A, 32           ; Otherwise load space
            CALL      PRINTCHAR       ; And call PRINTCHAR to delete cursor
            JP        NEWLINE         ; Now next line which then returns
NONEWLINE:  CP        CR              ; If CR
            RET       Z               ; Do nothing
            PUSH      AF              ; Save character to print
            CP        BKSP            ; Is it a backspace?
            JP        Z, BACKSP       ; Then do that
            LD        A, (CHARS)      ; Load character count
            CP        SCREENWIDTH     ; Maximum reached?
            CALL      Z, NEWLINE      ; Then next line
            
            LD        A, (CHARS)      ; Again the character count
            LD        H, A            ; Prepare multiply
            LD        E, 8            ; 8 pixel per char
            CALL      Mult8           ; HL = 8 * (CHARS)
            PUSH      HL
            POP       IX              ; IX = HL        
            
            LD        A, (LINES)      ; Now the same with the line count
            LD        H, A            ; Prepare multiply
            LD        E, 8            ; 8 pixel per line
            CALL      Mult8           ; HL = 8 * (LINES)
            PUSH      HL
            POP       IY              ; IY = HL
            
            LD        A, (CHARS)      ; Again the character count
            INC       A               ; Add one character
            LD        (CHARS), A      ; And save
            
            POP       AF
            LD        H, A
            LD        E, 8
            CALL      Mult8           ; Get address of ASCII char in font
            LD        BC, font  
            ADD       HL, BC          ; Then add the starting address of font
            LD        E, 8            ; E is the line counter
NEXTLINE:   LD        (COORDX), IX    ; Write to mem for GOTOCOORD
            LD        (COORDY), IY    
            CALL      GOTOCOORD       ; Set TFT to coord (IX, IY)
            LD        B, 8            ; Start bit with 7
            LD        C, 80H          ; Inverse counter
LINEOUT:    LD        A, (HL)         ; Load Font+8*Char+Line
            AND       C               ; Extract bit number B
            PUSH      BC
            JR        Z, SKIPPIX      ; Don't set pixel if bit not set
            LD				BC, (TXTCOL)        
            CALL			WRDATA					; Write Data to current register
            JR        NOSKIP
SKIPPIX:    LD				BC, BLACK
            CALL			WRDATA					; Write Data to current register
NOSKIP:     POP       BC
            SRL       C
            DJNZ      LINEOUT         ; Next pixel if line not finished
            DEC       E               ; Decrease line counter
            RET       Z               ; If 0 all lines done
            INC       IY              ; Otherwise next line
            INC       HL              ; Also next line (byte) in font
            JR        NEXTLINE        ; And go  
          
; ****************  TFT to coordinate (COORDX), (COORDY) ****************

GOTOCOORD:  LD        BC, 020H
            CALL      WRCMD
            LD        BC, (COORDX)
            CALL      WRDATA
            LD        BC, 021H
            CALL      WRCMD
            LD        BC, (COORDY)
            CALL      WRDATA
            LD				BC, 0022H
            CALL			WRCMD
            RET
          
; ******************            Start new line         ******************
            
NEWLINE:    PUSH      AF
            PUSH      BC
            PUSH      DE
            PUSH      HL
						
						LD        A, (LINES)      ; Load current number of lines
            INC       A
            CP        SCREENHEIGHT    ; Maximum reached?
            JR        NZ, SKPLNMAX    ; If not skip clear
            LD        A, 0            ; If yes set line to 0
SKPLNMAX:   LD        (LINES), A      ; Save current line
            LD        A, 0            
            LD        (CHARS), A      ; CR
						
						LD        BC, 006AH       ; Register for offset
						CALL      WRCMD
					  LD        A, (SCRNOFFSET) ; Now load the current offset
					  INC       A               ; Increase offset about one character line
						CP        SCREENHEIGHT    ; Max reached?
						JP        NZ, SKIPMAX     ; If not skip
						LD        A, 0            ; Start at top
SKIPMAX:    LD        (SCRNOFFSET), A ; And save
            LD        H, A
            LD        E, 8
            CALL      Mult8           ; HL = SCRNOFFSET * 8
						PUSH      HL
						POP       BC
						CALL      WRDATA          ; Set
					  LD        BC, 0022H       ; Default mode is writing RAM
            CALL			WRCMD
            
            LD        A, (LINES)      ; Load characters lines
            LD        H, A            ; Prepare multiply
            LD        E, 8            ; 8 pixel per line
            CALL      Mult8           ; HL = 8 * (LINES)
            LD        (COORDY), HL
            LD        A, 0
            LD        (COORDX), A
            CALL      GOTOCOORD       ; Set to first pixel of first line in current
                                      ; character line
            LD        A, 8            ; Number of lines to clear
            LD				DE, BLACK       ; Clear with black
            LD				C, LCDDATA      ; Port number needed in C                           
LOOPXNL:    LD        B, SCREENWIDTH*8
LOOPYNL:    OUT				(C), D
						OUT				(C), E
						DJNZ			LOOPYNL
						DEC       A
						JR        NZ, LOOPXNL
						CALL      GOTOCOORD
						
            POP       HL
            POP       DE
            POP       BC
            POP       AF
            RET

; *******************           Clear screen            ******************  

CLEAR:      PUSH      AF    
            PUSH      BC
            PUSH      DE
            PUSH      HL
            LD				A, 0
 						LD				(COORDX), A
 						LD				(COORDY), A
 						CALL			GOTOCOORD       ; Set cursor to beginning of screen
 						LD        HL, 320         ; Number of lines to clear
        		LD				DE, BLACK       ; Clear with black
 						LD				C, LCDDATA      ; Port number needed in C                           
LOOPX:      LD        B, SCREENWIDTH*8
LOOPY:      OUT				(C), D
						OUT				(C), E
						DJNZ			LOOPY
            DEC				HL 
            LD				A, H
						OR				L
            JR        NZ, LOOPX  
            POP       HL
            POP       DE
            POP       BC
            POP       AF
            RET
            
            
BACKSP:     LD        A, (CHARS)      ; Load character count
            CP        0               ; Does the line contain a character?
            JP        Z, BACKSPEND    ; Then do nothing
            LD        A, 32           ; Load space
            CALL      SAFEPRINT       ; Clear cursor
            LD        A, (CHARS)
            DEC       A
            DEC       A
            LD        (CHARS), A      ; Write count back
            LD				A, 32           ; Load space
            CALL      SAFEPRINT       ; Clear deleted char
            LD				A, (CHARS)      ; Decrement chars counter again
            DEC       A
            LD        (CHARS), A
BACKSPEND:  POP        AF
            RET

READKEYINIT:
            SETCOLOR(BLUE)
            PUSH      BC
            PUSH      DE
            PUSH      HL
            LD        E, 0            ; E will be the last pressed key
READKEY:    LD				HL, (CRSRSTATE) ; Load current cursor state
            LD        BC, 010000
						SBC       HL, BC          ; Cursor on at 10000
						JR				Z, SETCRSR      ; Then jump to set the cursor
						LD				HL, (CRSRSTATE) ; Load current cursor state
						LD        BC, 020000
						SBC       HL, BC          ; Cursor off at 20000
						JR        Z, NOCRSR       ; Then jump to it
						JR        SKIPSET         ; Not 10000 or 20000? Then do nothing
NOCRSR:     LD        A, 32           ; Load space
            LD        HL, 0           ; Reset counter
            LD        (CRSRSTATE), HL
            JR        WRITECRSR       ; And goto write       
SETCRSR:		LD        A, 219					; Cursor
WRITECRSR:  CALL      SAFEPRINT		
						LD				A, (CHARS)      ; Load number of characters
						DEC				A               ; Remove the cursor from character count
						LD        (CHARS), A      ; And save	
SKIPSET:    LD				HL, (CRSRSTATE)
            INC       HL
						LD        (CRSRSTATE), HL ; Safe for later check	
						LD        H, 2            ; H is the line register, start with second
            LD        B, 0            ; Count lines for later multiplication  
            LD        D, 0            ; DE will be the adress for mask
            
NEXTKEY:    LD        A, H            
            CP        0                ; All lines tried? 
            JP        Z, KEYOUT        ; Then check if there was a key pressed
            OUT       (REGISTER), A    ; Put current line to register
            IN        A, (KEYREAD)     ; Input Keys
            SLA       H                ; Next line
            INC       B
            CP        0                ; Was key zero?
            JP        Z, NEXTKEY       ; Then try again with next lines
            LD        D, 0             ; In D will be the number of the key
LOGARITHM:  INC       D                ; Add one per shift
            SRL       A                ; Shift key right
            JP        NZ, LOGARITHM    ; If not zero shift again
            DEC       D                ; Was too much
            LD        A, 1             ; Check first line for alt, shift, etc...
            OUT       (REGISTER), A
            IN        A, (KEYREAD)
            AND       1                ; Check if first bit set (shift key pressed)
            JP        NZ, LOADSHIFT    ; Then jump to read with shift
            LD        A, D             ; Put read key into accu
            ADD       A, KEYMAP        ; Add base of key map array
            JP        ADDOFFSET        ; Jump to load key
LOADSHIFT:  LD        A, D
            ADD       A, SHIFTKEYMAP   ; In this case add the base for shift    
ADDOFFSET:  ADD       A, 8             ; Add 8 for every line
            DJNZ      ADDOFFSET        ; Jump back (do while loop)
            SUB       8                ; Since do while is one too much
TRANSKEY:   XOR       B                ; Empty B
            LD        C, A             ; A will be address in BC
            LD        A, (BC)          ; Load key
            CP        E                ; Same key?
            JP        Z, READKEY       ; Then from beginning
            LD        E, A             ; Otherwise save new key
            JP        READKEY          ; And restart
KEYOUT:     LD        A, E
            LD        E, 0             ; empty it
            OR        A                ; Was a key read?
            JP        Z, READKEY       ; If not restart
            POP       HL
            POP       DE
            POP       BC
            RET  

Mult16:     LD        A, B
            LD        B, 16
MULT16LOOP: ADD       HL, HL
            SLA       C
            RLA
            JR        NC, MULT16SKIP
            ADD       HL, DE
MULT16SKIP: DJNZ      MULT16LOOP
            RET
  
Mult8:      LD        D, 0
            LD        L, D
            LD        B, 8
MULT8LOOP:  ADD       HL, HL
            JR        NC, MULT8SKIP
            ADD       HL, DE
MULT8SKIP:  DJNZ      MULT8LOOP
            RET

            
#include "font.asm"             
          

.end
